# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
from grpc_testing import _common
from grpc_testing._channel import _invocation


# All per-call credentials parameters are unused by this test infrastructure.
# pylint: disable=unused-argument
class UnaryUnary(grpc.UnaryUnaryMultiCallable):

    def __init__(self, method_full_rpc_name, channel_handler):
        self._method_full_rpc_name = method_full_rpc_name
        self._channel_handler = channel_handler

    def __call__(self, request, timeout=None, metadata=None, credentials=None):
        rpc_handler = self._channel_handler.invoke_rpc(
            self._method_full_rpc_name, _common.fuss_with_metadata(metadata),
            [request], True, timeout)
        return _invocation.blocking_unary_response(rpc_handler)

    def with_call(self, request, timeout=None, metadata=None, credentials=None):
        rpc_handler = self._channel_handler.invoke_rpc(
            self._method_full_rpc_name, _common.fuss_with_metadata(metadata),
            [request], True, timeout)
        return _invocation.blocking_unary_response_with_call(rpc_handler)

    def future(self, request, timeout=None, metadata=None, credentials=None):
        rpc_handler = self._channel_handler.invoke_rpc(
            self._method_full_rpc_name, _common.fuss_with_metadata(metadata),
            [request], True, timeout)
        return _invocation.future_call(rpc_handler)


class UnaryStream(grpc.StreamStreamMultiCallable):

    def __init__(self, method_full_rpc_name, channel_handler):
        self._method_full_rpc_name = method_full_rpc_name
        self._channel_handler = channel_handler

    def __call__(self, request, timeout=None, metadata=None, credentials=None):
        rpc_handler = self._channel_handler.invoke_rpc(
            self._method_full_rpc_name, _common.fuss_with_metadata(metadata),
            [request], True, timeout)
        return _invocation.ResponseIteratorCall(rpc_handler)


class StreamUnary(grpc.StreamUnaryMultiCallable):

    def __init__(self, method_full_rpc_name, channel_handler):
        self._method_full_rpc_name = method_full_rpc_name
        self._channel_handler = channel_handler

    def __call__(self,
                 request_iterator,
                 timeout=None,
                 metadata=None,
                 credentials=None):
        rpc_handler = self._channel_handler.invoke_rpc(
            self._method_full_rpc_name, _common.fuss_with_metadata(metadata),
            [], False, timeout)
        _invocation.consume_requests(request_iterator, rpc_handler)
        return _invocation.blocking_unary_response(rpc_handler)

    def with_call(self,
                  request_iterator,
                  timeout=None,
                  metadata=None,
                  credentials=None):
        rpc_handler = self._channel_handler.invoke_rpc(
            self._method_full_rpc_name, _common.fuss_with_metadata(metadata),
            [], False, timeout)
        _invocation.consume_requests(request_iterator, rpc_handler)
        return _invocation.blocking_unary_response_with_call(rpc_handler)

    def future(self,
               request_iterator,
               timeout=None,
               metadata=None,
               credentials=None):
        rpc_handler = self._channel_handler.invoke_rpc(
            self._method_full_rpc_name, _common.fuss_with_metadata(metadata),
            [], False, timeout)
        _invocation.consume_requests(request_iterator, rpc_handler)
        return _invocation.future_call(rpc_handler)


class StreamStream(grpc.StreamStreamMultiCallable):

    def __init__(self, method_full_rpc_name, channel_handler):
        self._method_full_rpc_name = method_full_rpc_name
        self._channel_handler = channel_handler

    def __call__(self,
                 request_iterator,
                 timeout=None,
                 metadata=None,
                 credentials=None):
        rpc_handler = self._channel_handler.invoke_rpc(
            self._method_full_rpc_name, _common.fuss_with_metadata(metadata),
            [], False, timeout)
        _invocation.consume_requests(request_iterator, rpc_handler)
        return _invocation.ResponseIteratorCall(rpc_handler)


# pylint: enable=unused-argument



# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc_testing


class UnaryUnaryServerRpc(grpc_testing.UnaryUnaryServerRpc):

    def __init__(self, handler):
        self._handler = handler

    def initial_metadata(self):
        return self._handler.initial_metadata()

    def cancel(self):
        self._handler.cancel()

    def termination(self):
        return self._handler.unary_response_termination()


class UnaryStreamServerRpc(grpc_testing.UnaryStreamServerRpc):

    def __init__(self, handler):
        self._handler = handler

    def initial_metadata(self):
        return self._handler.initial_metadata()

    def take_response(self):
        return self._handler.take_response()

    def cancel(self):
        self._handler.cancel()

    def termination(self):
        return self._handler.stream_response_termination()


class StreamUnaryServerRpc(grpc_testing.StreamUnaryServerRpc):

    def __init__(self, handler):
        self._handler = handler

    def initial_metadata(self):
        return self._handler.initial_metadata()

    def send_request(self, request):
        self._handler.add_request(request)

    def requests_closed(self):
        self._handler.requests_closed()

    def cancel(self):
        self._handler.cancel()

    def termination(self):
        return self._handler.unary_response_termination()


class StreamStreamServerRpc(grpc_testing.StreamStreamServerRpc):

    def __init__(self, handler):
        self._handler = handler

    def initial_metadata(self):
        return self._handler.initial_metadata()

    def send_request(self, request):
        self._handler.add_request(request)

    def requests_closed(self):
        self._handler.requests_closed()

    def take_response(self):
        return self._handler.take_response()

    def cancel(self):
        self._handler.cancel()

    def termination(self):
        return self._handler.stream_response_termination()



# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import threading

import grpc
from grpc_testing import _common

_CLIENT_INACTIVE = object()


class Handler(_common.ServerRpcHandler):

    @abc.abstractmethod
    def initial_metadata(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def add_request(self, request):
        raise NotImplementedError()

    @abc.abstractmethod
    def take_response(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def requests_closed(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def cancel(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def unary_response_termination(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def stream_response_termination(self):
        raise NotImplementedError()


class _Handler(Handler):

    def __init__(self, requests_closed):
        self._condition = threading.Condition()
        self._requests = []
        self._requests_closed = requests_closed
        self._initial_metadata = None
        self._responses = []
        self._trailing_metadata = None
        self._code = None
        self._details = None
        self._unary_response = None
        self._expiration_future = None
        self._termination_callbacks = []

    def send_initial_metadata(self, initial_metadata):
        with self._condition:
            self._initial_metadata = initial_metadata
            self._condition.notify_all()

    def take_request(self):
        with self._condition:
            while True:
                if self._code is None:
                    if self._requests:
                        request = self._requests.pop(0)
                        self._condition.notify_all()
                        return _common.ServerRpcRead(request, False, False)
                    elif self._requests_closed:
                        return _common.REQUESTS_CLOSED
                    else:
                        self._condition.wait()
                else:
                    return _common.TERMINATED

    def is_active(self):
        with self._condition:
            return self._code is None

    def add_response(self, response):
        with self._condition:
            self._responses.append(response)
            self._condition.notify_all()

    def send_termination(self, trailing_metadata, code, details):
        with self._condition:
            self._trailing_metadata = trailing_metadata
            self._code = code
            self._details = details
            if self._expiration_future is not None:
                self._expiration_future.cancel()
            self._condition.notify_all()

    def add_termination_callback(self, callback):
        with self._condition:
            if self._code is None:
                self._termination_callbacks.append(callback)
                return True
            else:
                return False

    def initial_metadata(self):
        with self._condition:
            while True:
                if self._initial_metadata is None:
                    if self._code is None:
                        self._condition.wait()
                    else:
                        raise ValueError(
                            'No initial metadata despite status code!')
                else:
                    return self._initial_metadata

    def add_request(self, request):
        with self._condition:
            self._requests.append(request)
            self._condition.notify_all()

    def take_response(self):
        with self._condition:
            while True:
                if self._responses:
                    response = self._responses.pop(0)
                    self._condition.notify_all()
                    return response
                elif self._code is None:
                    self._condition.wait()
                else:
                    raise ValueError('No more responses!')

    def requests_closed(self):
        with self._condition:
            self._requests_closed = True
            self._condition.notify_all()

    def cancel(self):
        with self._condition:
            if self._code is None:
                self._code = _CLIENT_INACTIVE
                termination_callbacks = self._termination_callbacks
                self._termination_callbacks = None
                if self._expiration_future is not None:
                    self._expiration_future.cancel()
                self._condition.notify_all()
        for termination_callback in termination_callbacks:
            termination_callback()

    def unary_response_termination(self):
        with self._condition:
            while True:
                if self._code is _CLIENT_INACTIVE:
                    raise ValueError('Huh? Cancelled but wanting status?')
                elif self._code is None:
                    self._condition.wait()
                else:
                    if self._unary_response is None:
                        if self._responses:
                            self._unary_response = self._responses.pop(0)
                    return (
                        self._unary_response,
                        self._trailing_metadata,
                        self._code,
                        self._details,
                    )

    def stream_response_termination(self):
        with self._condition:
            while True:
                if self._code is _CLIENT_INACTIVE:
                    raise ValueError('Huh? Cancelled but wanting status?')
                elif self._code is None:
                    self._condition.wait()
                else:
                    return self._trailing_metadata, self._code, self._details

    def expire(self):
        with self._condition:
            if self._code is None:
                if self._initial_metadata is None:
                    self._initial_metadata = _common.FUSSED_EMPTY_METADATA
                self._trailing_metadata = _common.FUSSED_EMPTY_METADATA
                self._code = grpc.StatusCode.DEADLINE_EXCEEDED
                self._details = 'Took too much time!'
                termination_callbacks = self._termination_callbacks
                self._termination_callbacks = None
                self._condition.notify_all()
        for termination_callback in termination_callbacks:
            termination_callback()

    def set_expiration_future(self, expiration_future):
        with self._condition:
            self._expiration_future = expiration_future


def handler_without_deadline(requests_closed):
    return _Handler(requests_closed)


def handler_with_deadline(requests_closed, time, deadline):
    handler = _Handler(requests_closed)
    expiration_future = time.call_at(handler.expire, deadline)
    handler.set_expiration_future(expiration_future)
    return handler



# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import threading

from grpc_testing import _common
from grpc_testing._channel import _rpc_state


class State(_common.ChannelHandler):

    def __init__(self):
        self._condition = threading.Condition()
        self._rpc_states = collections.defaultdict(list)

    def invoke_rpc(self, method_full_rpc_name, invocation_metadata, requests,
                   requests_closed, timeout):
        rpc_state = _rpc_state.State(invocation_metadata, requests,
                                     requests_closed)
        with self._condition:
            self._rpc_states[method_full_rpc_name].append(rpc_state)
            self._condition.notify_all()
        return rpc_state

    def take_rpc_state(self, method_descriptor):
        method_full_rpc_name = '/{}/{}'.format(
            method_descriptor.containing_service.full_name,
            method_descriptor.name)
        with self._condition:
            while True:
                method_rpc_states = self._rpc_states[method_full_rpc_name]
                if method_rpc_states:
                    return method_rpc_states.pop(0)
                else:
                    self._condition.wait()



# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Objects for use in testing gRPC Python-using application code."""

import abc
import six

from google.protobuf import descriptor

import grpc


class UnaryUnaryChannelRpc(six.with_metaclass(abc.ABCMeta)):
    """Fixture for a unary-unary RPC invoked by a system under test.

    Enables users to "play server" for the RPC.
    """

    @abc.abstractmethod
    def send_initial_metadata(self, initial_metadata):
        """Sends the RPC's initial metadata to the system under test.

        Args:
          initial_metadata: The RPC's initial metadata to be "sent" to
            the system under test.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def cancelled(self):
        """Blocks until the system under test has cancelled the RPC."""
        raise NotImplementedError()

    @abc.abstractmethod
    def terminate(self, response, trailing_metadata, code, details):
        """Terminates the RPC.

        Args:
          response: The response for the RPC.
          trailing_metadata: The RPC's trailing metadata.
          code: The RPC's status code.
          details: The RPC's status details.
        """
        raise NotImplementedError()


class UnaryStreamChannelRpc(six.with_metaclass(abc.ABCMeta)):
    """Fixture for a unary-stream RPC invoked by a system under test.

    Enables users to "play server" for the RPC.
    """

    @abc.abstractmethod
    def send_initial_metadata(self, initial_metadata):
        """Sends the RPC's initial metadata to the system under test.

        Args:
          initial_metadata: The RPC's initial metadata to be "sent" to
            the system under test.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def send_response(self, response):
        """Sends a response to the system under test.

        Args:
          response: A response message to be "sent" to the system under test.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def cancelled(self):
        """Blocks until the system under test has cancelled the RPC."""
        raise NotImplementedError()

    @abc.abstractmethod
    def terminate(self, trailing_metadata, code, details):
        """Terminates the RPC.

        Args:
          trailing_metadata: The RPC's trailing metadata.
          code: The RPC's status code.
          details: The RPC's status details.
        """
        raise NotImplementedError()


class StreamUnaryChannelRpc(six.with_metaclass(abc.ABCMeta)):
    """Fixture for a stream-unary RPC invoked by a system under test.

    Enables users to "play server" for the RPC.
    """

    @abc.abstractmethod
    def send_initial_metadata(self, initial_metadata):
        """Sends the RPC's initial metadata to the system under test.

        Args:
          initial_metadata: The RPC's initial metadata to be "sent" to
            the system under test.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def take_request(self):
        """Draws one of the requests added to the RPC by the system under test.

        This method blocks until the system under test has added to the RPC
        the request to be returned.

        Successive calls to this method return requests in the same order in
        which the system under test added them to the RPC.

        Returns:
          A request message added to the RPC by the system under test.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def requests_closed(self):
        """Blocks until the system under test has closed the request stream."""
        raise NotImplementedError()

    @abc.abstractmethod
    def cancelled(self):
        """Blocks until the system under test has cancelled the RPC."""
        raise NotImplementedError()

    @abc.abstractmethod
    def terminate(self, response, trailing_metadata, code, details):
        """Terminates the RPC.

        Args:
          response: The response for the RPC.
          trailing_metadata: The RPC's trailing metadata.
          code: The RPC's status code.
          details: The RPC's status details.
        """
        raise NotImplementedError()


class StreamStreamChannelRpc(six.with_metaclass(abc.ABCMeta)):
    """Fixture for a stream-stream RPC invoked by a system under test.

    Enables users to "play server" for the RPC.
    """

    @abc.abstractmethod
    def send_initial_metadata(self, initial_metadata):
        """Sends the RPC's initial metadata to the system under test.

        Args:
          initial_metadata: The RPC's initial metadata to be "sent" to the
            system under test.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def take_request(self):
        """Draws one of the requests added to the RPC by the system under test.

        This method blocks until the system under test has added to the RPC
        the request to be returned.

        Successive calls to this method return requests in the same order in
        which the system under test added them to the RPC.

        Returns:
          A request message added to the RPC by the system under test.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def send_response(self, response):
        """Sends a response to the system under test.

        Args:
          response: A response messages to be "sent" to the system under test.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def requests_closed(self):
        """Blocks until the system under test has closed the request stream."""
        raise NotImplementedError()

    @abc.abstractmethod
    def cancelled(self):
        """Blocks until the system under test has cancelled the RPC."""
        raise NotImplementedError()

    @abc.abstractmethod
    def terminate(self, trailing_metadata, code, details):
        """Terminates the RPC.

        Args:
          trailing_metadata: The RPC's trailing metadata.
          code: The RPC's status code.
          details: The RPC's status details.
        """
        raise NotImplementedError()


class Channel(six.with_metaclass(abc.ABCMeta, grpc.Channel)):
    """A grpc.Channel double with which to test a system that invokes RPCs."""

    @abc.abstractmethod
    def take_unary_unary(self, method_descriptor):
        """Draws an RPC currently being made by the system under test.

        If the given descriptor does not identify any RPC currently being made
        by the system under test, this method blocks until the system under
        test invokes such an RPC.

        Args:
          method_descriptor: A descriptor.MethodDescriptor describing a
            unary-unary RPC method.

        Returns:
          A (invocation_metadata, request, unary_unary_channel_rpc) tuple of
            the RPC's invocation metadata, its request, and a
            UnaryUnaryChannelRpc with which to "play server" for the RPC.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def take_unary_stream(self, method_descriptor):
        """Draws an RPC currently being made by the system under test.

        If the given descriptor does not identify any RPC currently being made
        by the system under test, this method blocks until the system under
        test invokes such an RPC.

        Args:
          method_descriptor: A descriptor.MethodDescriptor describing a
            unary-stream RPC method.

        Returns:
          A (invocation_metadata, request, unary_stream_channel_rpc) tuple of
            the RPC's invocation metadata, its request, and a
            UnaryStreamChannelRpc with which to "play server" for the RPC.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def take_stream_unary(self, method_descriptor):
        """Draws an RPC currently being made by the system under test.

        If the given descriptor does not identify any RPC currently being made
        by the system under test, this method blocks until the system under
        test invokes such an RPC.

        Args:
          method_descriptor: A descriptor.MethodDescriptor describing a
            stream-unary RPC method.

        Returns:
          A (invocation_metadata, stream_unary_channel_rpc) tuple of the RPC's
            invocation metadata and a StreamUnaryChannelRpc with which to "play
            server" for the RPC.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def take_stream_stream(self, method_descriptor):
        """Draws an RPC currently being made by the system under test.

        If the given descriptor does not identify any RPC currently being made
        by the system under test, this method blocks until the system under
        test invokes such an RPC.

        Args:
          method_descriptor: A descriptor.MethodDescriptor describing a
            stream-stream RPC method.

        Returns:
          A (invocation_metadata, stream_stream_channel_rpc) tuple of the RPC's
            invocation metadata and a StreamStreamChannelRpc with which to
            "play server" for the RPC.
        """
        raise NotImplementedError()


class UnaryUnaryServerRpc(six.with_metaclass(abc.ABCMeta)):
    """Fixture for a unary-unary RPC serviced by a system under test.

    Enables users to "play client" for the RPC.
    """

    @abc.abstractmethod
    def initial_metadata(self):
        """Accesses the initial metadata emitted by the system under test.

        This method blocks until the system under test has added initial
        metadata to the RPC (or has provided one or more response messages or
        has terminated the RPC, either of which will cause gRPC Python to
        synthesize initial metadata for the RPC).

        Returns:
          The initial metadata for the RPC.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def cancel(self):
        """Cancels the RPC."""
        raise NotImplementedError()

    @abc.abstractmethod
    def termination(self):
        """Blocks until the system under test has terminated the RPC.

        Returns:
          A (response, trailing_metadata, code, details) sequence with the RPC's
            response, trailing metadata, code, and details.
        """
        raise NotImplementedError()


class UnaryStreamServerRpc(six.with_metaclass(abc.ABCMeta)):
    """Fixture for a unary-stream RPC serviced by a system under test.

    Enables users to "play client" for the RPC.
    """

    @abc.abstractmethod
    def initial_metadata(self):
        """Accesses the initial metadata emitted by the system under test.

        This method blocks until the system under test has added initial
        metadata to the RPC (or has provided one or more response messages or
        has terminated the RPC, either of which will cause gRPC Python to
        synthesize initial metadata for the RPC).

        Returns:
          The initial metadata for the RPC.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def take_response(self):
        """Draws one of the responses added to the RPC by the system under test.

        Successive calls to this method return responses in the same order in
        which the system under test added them to the RPC.

        Returns:
          A response message added to the RPC by the system under test.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def cancel(self):
        """Cancels the RPC."""
        raise NotImplementedError()

    @abc.abstractmethod
    def termination(self):
        """Blocks until the system under test has terminated the RPC.

        Returns:
          A (trailing_metadata, code, details) sequence with the RPC's trailing
            metadata, code, and details.
        """
        raise NotImplementedError()


class StreamUnaryServerRpc(six.with_metaclass(abc.ABCMeta)):
    """Fixture for a stream-unary RPC serviced by a system under test.

    Enables users to "play client" for the RPC.
    """

    @abc.abstractmethod
    def initial_metadata(self):
        """Accesses the initial metadata emitted by the system under test.

        This method blocks until the system under test has added initial
        metadata to the RPC (or has provided one or more response messages or
        has terminated the RPC, either of which will cause gRPC Python to
        synthesize initial metadata for the RPC).

        Returns:
          The initial metadata for the RPC.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def send_request(self, request):
        """Sends a request to the system under test.

        Args:
          request: A request message for the RPC to be "sent" to the system
            under test.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def requests_closed(self):
        """Indicates the end of the RPC's request stream."""
        raise NotImplementedError()

    @abc.abstractmethod
    def cancel(self):
        """Cancels the RPC."""
        raise NotImplementedError()

    @abc.abstractmethod
    def termination(self):
        """Blocks until the system under test has terminated the RPC.

        Returns:
          A (response, trailing_metadata, code, details) sequence with the RPC's
            response, trailing metadata, code, and details.
        """
        raise NotImplementedError()


class StreamStreamServerRpc(six.with_metaclass(abc.ABCMeta)):
    """Fixture for a stream-stream RPC serviced by a system under test.

    Enables users to "play client" for the RPC.
    """

    @abc.abstractmethod
    def initial_metadata(self):
        """Accesses the initial metadata emitted by the system under test.

        This method blocks until the system under test has added initial
        metadata to the RPC (or has provided one or more response messages or
        has terminated the RPC, either of which will cause gRPC Python to
        synthesize initial metadata for the RPC).

        Returns:
          The initial metadata for the RPC.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def send_request(self, request):
        """Sends a request to the system under test.

        Args:
          request: A request message for the RPC to be "sent" to the system
            under test.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def requests_closed(self):
        """Indicates the end of the RPC's request stream."""
        raise NotImplementedError()

    @abc.abstractmethod
    def take_response(self):
        """Draws one of the responses added to the RPC by the system under test.

        Successive calls to this method return responses in the same order in
        which the system under test added them to the RPC.

        Returns:
          A response message added to the RPC by the system under test.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def cancel(self):
        """Cancels the RPC."""
        raise NotImplementedError()

    @abc.abstractmethod
    def termination(self):
        """Blocks until the system under test has terminated the RPC.

        Returns:
          A (trailing_metadata, code, details) sequence with the RPC's trailing
            metadata, code, and details.
        """
        raise NotImplementedError()


class Server(six.with_metaclass(abc.ABCMeta)):
    """A server with which to test a system that services RPCs."""

    @abc.abstractmethod
    def invoke_unary_unary(self, method_descriptor, invocation_metadata,
                           request, timeout):
        """Invokes an RPC to be serviced by the system under test.

        Args:
          method_descriptor: A descriptor.MethodDescriptor describing a unary-unary
            RPC method.
          invocation_metadata: The RPC's invocation metadata.
          request: The RPC's request.
          timeout: A duration of time in seconds for the RPC or None to
            indicate that the RPC has no time limit.

        Returns:
          A UnaryUnaryServerRpc with which to "play client" for the RPC.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def invoke_unary_stream(self, method_descriptor, invocation_metadata,
                            request, timeout):
        """Invokes an RPC to be serviced by the system under test.

        Args:
          method_descriptor: A descriptor.MethodDescriptor describing a unary-stream
            RPC method.
          invocation_metadata: The RPC's invocation metadata.
          request: The RPC's request.
          timeout: A duration of time in seconds for the RPC or None to
            indicate that the RPC has no time limit.

        Returns:
          A UnaryStreamServerRpc with which to "play client" for the RPC.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def invoke_stream_unary(self, method_descriptor, invocation_metadata,
                            timeout):
        """Invokes an RPC to be serviced by the system under test.

        Args:
          method_descriptor: A descriptor.MethodDescriptor describing a stream-unary
            RPC method.
          invocation_metadata: The RPC's invocation metadata.
          timeout: A duration of time in seconds for the RPC or None to
            indicate that the RPC has no time limit.

        Returns:
          A StreamUnaryServerRpc with which to "play client" for the RPC.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def invoke_stream_stream(self, method_descriptor, invocation_metadata,
                             timeout):
        """Invokes an RPC to be serviced by the system under test.

        Args:
          method_descriptor: A descriptor.MethodDescriptor describing a stream-stream
            RPC method.
          invocation_metadata: The RPC's invocation metadata.
          timeout: A duration of time in seconds for the RPC or None to
            indicate that the RPC has no time limit.

        Returns:
          A StreamStreamServerRpc with which to "play client" for the RPC.
        """
        raise NotImplementedError()


class Time(six.with_metaclass(abc.ABCMeta)):
    """A simulation of time.

    Implementations needn't be connected with real time as provided by the
    Python interpreter, but as long as systems under test use
    RpcContext.is_active and RpcContext.time_remaining for querying RPC liveness
    implementations may be used to change passage of time in tests.
    """

    @abc.abstractmethod
    def time(self):
        """Accesses the current test time.

        Returns:
          The current test time (over which this object has authority).
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def call_in(self, behavior, delay):
        """Adds a behavior to be called after some time.

        Args:
          behavior: A behavior to be called with no arguments.
          delay: A duration of time in seconds after which to call the behavior.

        Returns:
          A grpc.Future with which the call of the behavior may be cancelled
            before it is executed.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def call_at(self, behavior, time):
        """Adds a behavior to be called at a specific time.

        Args:
          behavior: A behavior to be called with no arguments.
          time: The test time at which to call the behavior.

        Returns:
          A grpc.Future with which the call of the behavior may be cancelled
            before it is executed.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def sleep_for(self, duration):
        """Blocks for some length of test time.

        Args:
          duration: A duration of test time in seconds for which to block.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def sleep_until(self, time):
        """Blocks until some test time.

        Args:
          time: The test time until which to block.
        """
        raise NotImplementedError()


def strict_real_time():
    """Creates a Time backed by the Python interpreter's time.

    The returned instance will be "strict" with respect to callbacks
    submitted to it: it will ensure that all callbacks registered to
    be called at time t have been called before it describes the time
    as having advanced beyond t.

    Returns:
      A Time backed by the "system" (Python interpreter's) time.
    """
    from grpc_testing import _time
    return _time.StrictRealTime()


def strict_fake_time(now):
    """Creates a Time that can be manipulated by test code.

    The returned instance maintains an internal representation of time
    independent of real time. This internal representation only advances
    when user code calls the instance's sleep_for and sleep_until methods.

    The returned instance will be "strict" with respect to callbacks
    submitted to it: it will ensure that all callbacks registered to
    be called at time t have been called before it describes the time
    as having advanced beyond t.

    Returns:
      A Time that simulates the passage of time.
    """
    from grpc_testing import _time
    return _time.StrictFakeTime(now)


def channel(service_descriptors, time):
    """Creates a Channel for use in tests of a gRPC Python-using system.

    Args:
      service_descriptors: An iterable of descriptor.ServiceDescriptors
        describing the RPCs that will be made on the returned Channel by the
        system under test.
      time: A Time to be used for tests.

    Returns:
      A Channel for use in tests.
    """
    from grpc_testing import _channel
    return _channel.testing_channel(service_descriptors, time)


def server_from_dictionary(descriptors_to_servicers, time):
    """Creates a Server for use in tests of a gRPC Python-using system.

    Args:
      descriptors_to_servicers: A dictionary from descriptor.ServiceDescriptors
        defining RPC services to servicer objects (usually instances of classes
        that implement "Servicer" interfaces defined in generated "_pb2_grpc"
        modules) implementing those services.
      time: A Time to be used for tests.

    Returns:
      A Server for use in tests.
    """
    from grpc_testing import _server
    return _server.server_from_dictionary(descriptors_to_servicers, time)
